#include "aspen_ddreal.h"
#include "aspen_ddcomplex.h"
#if TYPE_DOUBLE_COMPLEX || TYPE_FLOAT_COMPLEX
#include "aspen_complex.h"
#endif

#if TYPE_DD
#   define TYPE			cuddreal
#   define TYPE_		double
#   define CONST(v)		makeCONST<cuddreal>(v)
#   define Conj(v)		makeCONST<cuddreal>(v)
#   define __isDD__			1
#   define __isDOUBLE__			0
#   define __isFLOAT__			0
#   define __isDD_COMPLEX__		0
#   define __isDOUBLE_COMPLEX__		0
#   define __isFLOAT_COMPLEX__		0
#   define EPS			(3.1556e-30)
#   define _FUNCS		W_FUNCS
#endif

#if TYPE_DOUBLE
#   define TYPE			double
#   define TYPE_		double
#   define CONST(v)		((double)(v))
#   define Conj(v)		((double)(v))
#   define __isDD__			0
#   define __isDOUBLE__			1
#   define __isFLOAT__			0
#   define __isDD_COMPLEX__		0
#   define __isDOUBLE_COMPLEX__		0
#   define __isFLOAT_COMPLEX__		0
#   define EPS			(1.7764e-15)
#   define _FUNCS		D_FUNCS
#endif

#if TYPE_FLOAT
#   define TYPE			float
#   define TYPE_		float
#   define CONST(v)		((float)(v))
#   define Conj(v)		((float)(v))
#   define __isDD__			0
#   define __isDOUBLE__			0
#   define __isFLOAT__			1
#   define __isDD_COMPLEX__		0
#   define __isDOUBLE_COMPLEX__		0
#   define __isFLOAT_COMPLEX__		0
#   define EPS			(1.1921e-07)
#   define _FUNCS		S_FUNCS
#endif

#if TYPE_DD_COMPLEX
#   define TYPE			cuddcomplex
#   define TYPE_		cuddreal
#   define CONST(v)		makeCONST<cuddcomplex>(v)
#   define __isDD__			0
#   define __isDOUBLE__			0
#   define __isFLOAT__			0
#   define __isDD_COMPLEX__		1
#   define __isDOUBLE_COMPLEX__		0
#   define __isFLOAT_COMPLEX__		0
#   define EPS			(3.1556e-30)
#   define _FUNCS		U_FUNCS
#endif

#if TYPE_DOUBLE_COMPLEX
#   define TYPE			cuDoubleComplex
#   define TYPE_		double
#   define CONST(v)		make_cuDoubleComplex(v,(double)0)
#   define __isDD__			0
#   define __isDOUBLE__			0
#   define __isFLOAT__			0
#   define __isDD_COMPLEX__		0
#   define __isDOUBLE_COMPLEX__		1
#   define __isFLOAT_COMPLEX__		0
#   define EPS			(1.7764e-15)
#   define _FUNCS		Z_FUNCS
#endif

#if TYPE_FLOAT_COMPLEX
#   define TYPE			cuFloatComplex
#   define TYPE_		float
#   define CONST(v)		make_cuFloatComplex(v,(float)0)
#   define __isDD__			0
#   define __isDOUBLE__			0
#   define __isFLOAT__			0
#   define __isDD_COMPLEX__		0
#   define __isDOUBLE_COMPLEX__		0
#   define __isFLOAT_COMPLEX__		1
#   define EPS			(1.1921e-07)
#   define _FUNCS		C_FUNCS
#endif

// macro for the legacy cublas
#define	cublasAlloc( a, b, c )	cudaMalloc ( (c), size_t((a)*(b)) )
#define	cublasFree( a )		cudaFree ( a )



#ifdef __cplusplus
extern "C" {
#endif

  double get_wtime(void);
  int do_usleep(int usec);
  int get_gpu_id(void);
  void print_head(char *func_name, int argc, char **argv);

  void
  ASPEN_init( int );

  void
  ASPEN_shutdown( void );

  int
  ASPEN_get_driver_Version( void );

  int
  ASPEN_get_runtime_Version( void );

  size_t
  get_device_WorkSize( int id, int size );

  void
  ASPEN_SSYMVu(int blk, int N, float *dA, int LDA,
               float *dB, int incx,
               float *dC, int incy,
               float alpha, float beta);

  void
  ASPEN_SSYMVl(int blk, int N, float *dA, int LDA,
               float *dB, int incx,
               float *dC, int incy,
               float alpha, float beta);

  void
  ASPEN_DSYMVu(int blk, int N, double *dA, int LDA,
               double *dB, int incx,
               double *dC, int incy,
               double alpha, double beta);
  void
  ASPEN_DSYMVl(int blk, int N, double *dA, int LDA,
               double *dB, int incx,
               double *dC, int incy,
               double alpha, double beta);

  void
  ASPEN_WSYMVu(int blk, int N, cuddreal *dA, int LDA,
               cuddreal *dB, int incx,
               cuddreal *dC, int incy,
               cuddreal alpha, cuddreal beta);
  void
  ASPEN_WSYMVl(int blk, int N, cuddreal *dA, int LDA,
               cuddreal *dB, int incx,
               cuddreal *dC, int incy,
               cuddreal alpha, cuddreal beta);

  void
  ASPEN_CHEMVu(int blk, int N, cuFloatComplex *dA, int LDA,
               cuFloatComplex *dB, int incx,
               cuFloatComplex *dC, int incy,
               cuFloatComplex alpha, cuFloatComplex beta);
  void
  ASPEN_CHEMVl(int blk, int N, cuFloatComplex *dA, int LDA,
               cuFloatComplex *dB, int incx,
               cuFloatComplex *dC, int incy,
               cuFloatComplex alpha, cuFloatComplex beta);

  void
  ASPEN_ZHEMVu(int blk, int N, cuDoubleComplex *dA, int LDA,
               cuDoubleComplex *dB, int incx,
               cuDoubleComplex *dC, int incy,
               cuDoubleComplex alpha, cuDoubleComplex beta);
  void
  ASPEN_ZHEMVl(int blk, int N, cuDoubleComplex *dA, int LDA,
               cuDoubleComplex *dB, int incx,
               cuDoubleComplex *dC, int incy,
               cuDoubleComplex alpha, cuDoubleComplex beta);

  void
  ASPEN_UHEMVu(int blk, int N, cuddcomplex *dA, int LDA,
               cuddcomplex *dB, int incx,
               cuddcomplex *dC, int incy,
               cuddcomplex alpha, cuddcomplex beta);
  void
  ASPEN_UHEMVl(int blk, int N, cuddcomplex *dA, int LDA,
               cuddcomplex *dB, int incx,
               cuddcomplex *dC, int incy,
               cuddcomplex alpha, cuddcomplex beta);

#ifdef __cplusplus
}
#endif

enum S_FUNCS { SSYMV_U, SSYMV_L };
enum D_FUNCS { DSYMV_U, DSYMV_L };
enum W_FUNCS { WSYMV_U, WSYMV_L };
enum C_FUNCS { CHEMV_U, CHEMV_L };
enum Z_FUNCS { ZHEMV_U, ZHEMV_L };
enum U_FUNCS { UHEMV_U, UHEMV_L };

