#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include <complex.h>
#include <omp.h>
extern "C" {
#include <cblas.h>
}
#include <cuda_runtime.h>
#include <cublas_v2.h>

#define	ITER	6

#include "aspen_const.h"


#if TYPE_DD
#include "aspen_ddreal.h"
#endif
#if TYPE_DD_COMPLEX
#include "aspen_ddreal.h"
#include "aspen_ddcomplex.h"
#endif

#define USE_CUDA_EVENT_TIMER	0
#include "test-common.h"
#include "test.h"


int
set_matrix_vectors( const int N, TYPE *A, const int lda, TYPE *X, TYPE *Y,
                    const enum _FUNCS func );



int
set_matrix_vectors( const int N, TYPE *A, const int lda, TYPE *X, TYPE *Y,
                    const enum _FUNCS func )
{
#pragma omp parallel
  {
#pragma omp for
    for(int i=0;i<N;i++) {
      for(int j=0;j<N;j++) {
        A[j+i*lda]=makeCONST<TYPE>(0);
      }
    }

#pragma omp for
    for(int i=0;i<N;i++) X[i]=makeCONST<TYPE>(1);
#pragma omp for
    for(int i=0;i<N;i++) Y[i]=makeCONST<TYPE>(1);
  }
}


int
main(int argc, char *argv[])
{
  TYPE	alpha, beta;

  TYPE	*A,  *B,  *C;
  TYPE	*dA, *dB, *dC;

  int	N, M, K;
  int	LDA, LDB, LDC;
  int	i;
  int	blk;
  FILE	*fp;
  int	id = get_gpu_id();

  enum _FUNCS func;
  char	*func_name = (char *)"";


  //  CUDA device setup
  cudaError_t st;
  if ( cudaSuccess != (st = cudaSetDevice( id )) ) {
    exit(1);
  }

  cudaDeviceReset();
  cudaDeviceReset();

  if ( cudaSuccess != (st = cudaSetDevice( id )) ) {
    exit(1);
  }

  //  CUDA BLAS initialization
  ASPEN_init( id );


  //  function name check
  {
    char *argv0 = argv[0];
    if ( strrchr(argv0, '/') != argv0 ) {
      argv0 = strrchr(argv0,'/')+1;
    }
#if TYPE_DD || TYPE_DOUBLE || TYPE_FLOAT
    if ( 0 == strcmp(argv0,TEST2_SYMV_U) ) {
      func      = SYMV_U;
      func_name = (char *)NAME_SYMV_U;
    }
    if ( 0 == strcmp(argv0,TEST2_SYMV_L) ) {
      func      = SYMV_L;
      func_name = (char *)NAME_SYMV_L;
    }
#else
    if ( 0 == strcmp(argv0,TEST2_HEMV_U) ) {
      func      = HEMV_U;
      func_name = (char *)NAME_HEMV_U;
    }
    if ( 0 == strcmp(argv0,TEST2_HEMV_L) ) {
      func      = HEMV_L;
      func_name = (char *)NAME_HEMV_L;
    }
#endif
  }


  // arg list check
  if ( argc > 1 ) {
    fp=fopen(argv[1], "r");
  } else {
    fp=fopen("IN", "r");
  }
  if ( fp == NULL ) {
    fprintf(stderr,"Cannot open parameter file.\n");
    exit(1);
  } else {
    // dummy
    fscanf(fp,"%d", &N);
  }
  if ( argc > 2 ) {
    sscanf(argv[2], "%d", &blk);
    printf("% BLK specified=%d\n", blk);
  } else {
    blk = -1; //64;
  }


  // print header
  print_head( func_name, argc, argv );
  fflush(stdout);


  // Determining the maximum size of arrays
  int N_fault=0;
  int N_max = get_device_WorkSize( id, sizeof(TYPE) );
  printf("Nmax=%d\n", N_max);


  // Allocating the host data
  {
    int M = N = N_max;
    int	LDM = ((M-1)/(256/sizeof(TYPE))+1);
    LDM *= (256/sizeof(TYPE));

    size_t len;
    len = ((size_t)sizeof(TYPE)*M)*LDM;
    A = (TYPE *)malloc( len );
    len = ((size_t)sizeof(TYPE)*N);
    B = (TYPE *)malloc( len );
    len = ((size_t)sizeof(TYPE)*N);
    C = (TYPE *)malloc( len );
  }


  // setting scalar parameters
  alpha =  makeCONST<TYPE>(3);
  alpha += makeCONST<TYPE>(3*EPS);
  beta  =  makeCONST<TYPE>(4);
  beta  += makeCONST<TYPE>(4*EPS);



  for(int Itr = 0; fscanf(fp,"%d", &N) > 0; Itr++ ) {

    if ( N <= 1 || N > N_max ) break;


    int	M = N;

    if ( argc <= 3 )
      {
        if ( sizeof(TYPE) == 4 ) {
          if ( M < 12600 ) M = 12600;
        }
        if ( sizeof(TYPE) == 8 ) {
          if ( M < 9000 ) M = 9000;
        }
        if ( sizeof(TYPE) == 16 ) {
          if ( M < 6300 ) M = 6300;
        }
        if ( sizeof(TYPE) == 32 ) {
          if ( M < 4500 ) M = 4500;
        }
      }

    int	LDM = ((M-1)/(256/sizeof(TYPE))+1);
    LDM *= (256/sizeof(TYPE));
    LDA = ((N-1)/(256/sizeof(TYPE))+1);
    LDA *= (256/sizeof(TYPE));


    // Allocating the device data
    cublasAlloc( M*LDM, sizeof(TYPE), (void**)&dA );
    cublasAlloc( N*1,   sizeof(TYPE), (void**)&dB );
    cublasAlloc( N*1,   sizeof(TYPE), (void**)&dC );

    if(dA==NULL||dB==NULL||dC==NULL) {
      fprintf(stderr, "# Fail to allocate Dev_array %dx%d, %d*4\n",
              N, LDA, N);
      fprintf(stderr, "# Fail to allocate memory %x %x %x \n",
              dA, dB, dC);
      if(dA!=NULL) cublasFree(dA);
      if(dB!=NULL) cublasFree(dB);
      if(dC!=NULL) cublasFree(dC);
      N_fault++;
      goto FFF;
    } else {
      N_fault=0;
    }

    printf("N= %d ", N);
    fflush(stdout);
    fflush(stdout);
    fflush(stdout);


    set_matrix_vectors( N, A, LDA, B, C, func );

#if 1
    if ( argc > 3 ) { ; } else {
      cublasSetMatrix(LDM, M, sizeof(TYPE), A, LDM, dA, LDM);
      cublasSetVector(N, sizeof(TYPE), B, 1, dB, 1);
      cublasSetVector(N, sizeof(TYPE), C, 1, dC, 1);
    }
#endif

    fflush(stdout);
    cudaThreadSynchronize();

    double tt[ITER];
    for(int i=0;i<ITER;i++){

      cudaThreadSynchronize();
      double tt1 = get_wtime();

      int	incb = 1;
      int	incc = 1;
      char	*opt = (char *)"U";

      switch (func) {

#if TYPE_DD || TYPE_DOUBLE || TYPE_FLOAT
      case SYMV_U:
        ASPEN_SYMVu (blk, N, dA, LDA, dB, 1, dC, 1, alpha, beta);
        break;

      case SYMV_L:
        ASPEN_SYMVl (blk, N, dA, LDA, dB, 1, dC, 1, alpha, beta);
        break;
#endif

#if TYPE_DD_COMPLEX || TYPE_DOUBLE_COMPLEX || TYPE_FLOAT_COMPLEX
      case HEMV_U:
        ASPEN_HEMVu (blk, N, dA, LDA, dB, 1, dC, 1, alpha, beta);
        break;

      case HEMV_L:
        ASPEN_HEMVl (blk, N, dA, LDA, dB, 1, dC, 1, alpha, beta);
        break;
#endif

      }

      cudaThreadSynchronize();
      double tt2 = get_wtime();
      tt[i] = tt2 - tt1;

    }

  EEE: fflush(stdout);


    { // sorting
      for(int i=0;i<ITER;i++) 
        for(int j=i+1;j<ITER;j++)
          if(tt[i]>tt[j]){ double s=tt[i]; tt[i]=tt[j]; tt[j]=s; }

      // stats
      double	tmax=tt[ITER-1];
      double	t = tt[(ITER-1)/2];
      double	tvar=0.0;
      for(int i=0;i<ITER;i++) tvar+=(tt[i]-t)*(tt[i]-t);
      tvar-=(tmax-t)*(tmax-t);
      tvar=sqrt(tvar)/(ITER-1);

      fflush(stdout);
      printf("%g [s] ", t);
#if TYPE_DD || TYPE_DOUBLE || TYPE_FLOAT
      printf("%g GFLOPS ", ((2.0*N)*N/t*1e-9));
#else
      printf("%g GFLOPS ", (4*(2.0*N)*N/t*1e-9));
#endif
      printf("%g ", tvar);
      if ( tvar/t > 2E-2 ) {
        printf("**%g**", tvar/t );
      }
      printf("%d ", LDA);
      fflush(stdout);
    }



  GGG: fflush(stdout);

    cublasFree(dA); //printf(":dA");
    cublasFree(dB); //printf(":dB");
    cublasFree(dC); //printf(":dC");

  FFF: fflush(stdout);

    printf("\n");
    fflush(stdout);
    fflush(stdout);

    if ( N_fault > 4 ) break;
    if ( feof(fp) ) break;
  }

  {
    free(A); //printf(":A");
    free(B); //printf(":B");
    free(C); //printf(":C");
  }

  ASPEN_shutdown();
 HHH:
  cudaDeviceReset();
  fclose(fp);
 FIN:
  ;

}

