#ifndef	ASPEN_H_INCLUDED

#include "aspen_version.h"
#include "aspen_types.h"


#ifdef __cplusplus
extern "C" {
#endif
  
  
  /* User must call ASPEN_init for initialization */
  int
  ASPEN_init ( const int device_id );
  
  /* User must finalize the ASPEN library by ASPEN_shutdown */
  int
  ASPEN_shutdown ( void );

  /* Get version info ASPEN_get_version_info */
  void 
  ASPEN_get_version_info ( int &version, char *codename, char *releasedate );
  
  
  void
  ASPEN_ssymv ( const char uplo, const int n,
                const float alpha, 
                const float * a, const int lda,
                const float * x, const int incx,
                const float beta,  
                float * y, const int incy );
  void
  ASPEN_dsymv ( const char uplo, const int n,
                const double alpha, 
                const double * a, const int lda,
                const double * x, const int incx,
                const double beta,  
                double * y, const int incy );

  void
  ASPEN_chemv ( const char uplo, const int n,
                const cuFloatComplex alpha, 
                const cuFloatComplex * a, const int lda,
                const cuFloatComplex * x, const int incx,
                cuFloatComplex beta,  
                cuFloatComplex * y, const int incy );
  void
  ASPEN_zhemv ( const char uplo, const int n,
                const cuDoubleComplex alpha, 
                const cuDoubleComplex * a, const int lda,
                const cuDoubleComplex * x, const int incx,
                const cuDoubleComplex beta,  
                cuDoubleComplex * y, const int incy );

  void
  ASPEN_wsymv ( const char uplo, const int n,
                const cuddreal alpha, 
                const cuddreal * a, const int lda,
                const cuddreal * x, const int incx,
                const cuddreal beta,  
                cuddreal * y, const int incy );

  void
  ASPEN_uhemv ( const char uplo, const int n,
                const cuddcomplex alpha, 
                const cuddcomplex * a, const int lda,
                const cuddcomplex * x, const int incx,
                const cuddcomplex beta,  
                cuddcomplex * y, const int incy );



#if ASPEN_HALF_ENABLED
  void
  ASPEN_haxpy ( const int n,
                const half alpha,
                const half * x, const int incx,
                half * y, const int incy );
#endif
  void
  ASPEN_saxpy ( const int n,
                const float alpha,
                const float * x, const int incx,
                float * y, const int incy );
  void
  ASPEN_daxpy ( const int n,
                const double alpha,
                const double * x, const int incx,
                double * y, const int incy );
  void
  ASPEN_caxpy ( const int n,
                const cuFloatComplex alpha,
                const cuFloatComplex * x, const int incx,
                cuFloatComplex * y, const int incy );
  void
  ASPEN_zaxpy ( const int n,
                const cuDoubleComplex alpha,
                const cuDoubleComplex * x, const int incx,
                cuDoubleComplex * y, const int incy );
  void
  ASPEN_waxpy ( const int n,
                const cuddreal alpha,
                const cuddreal * x, const int incx,
                cuddreal * y, const int incy );
  void
  ASPEN_uaxpy ( const int n,
                const cuddcomplex alpha,
                const cuddcomplex * x, const int incx,
                cuddcomplex * y, const int incy );
 

#if ASPEN_HALF_ENABLED
  void
  ASPEN_haxpby ( const int n,
                 const half alpha,
                 const half * x, const int incx,
                 const half beta,
                 half * y, const int incy );
#endif
  void
  ASPEN_saxpby ( const int n,
                 const float alpha,
                 const float * x, const int incx,
                 const float beta,
                 float * y, const int incy );
  void
  ASPEN_daxpby ( const int n,
                 const double alpha,
                 const double * x, const int incx,
                 const double beta,
                 double * y, const int incy );
  void
  ASPEN_caxpby ( const int n,
                 const cuFloatComplex alpha,
                 const cuFloatComplex * x, const int incx,
                 const cuFloatComplex beta,
                 cuFloatComplex * y, const int incy );
  void
  ASPEN_zaxpby ( const int n,
                 const cuDoubleComplex alpha,
                 const cuDoubleComplex * x, const int incx,
                 const cuDoubleComplex beta,
                 cuDoubleComplex * y, const int incy );
  void
  ASPEN_waxpby ( const int n,
                 const cuddreal alpha,
                 const cuddreal * x, const int incx,
                 const cuddreal beta,
                 cuddreal * y, const int incy );
  void
  ASPEN_uaxpby ( const int n,
                 const cuddcomplex alpha,
                 const cuddcomplex * x, const int incx,
                 const cuddcomplex beta,
                 cuddcomplex * y, const int incy );


#if ASPEN_HALF_ENABLED
  void
  ASPEN_hswap ( const int n,
                half * x, const int incx,
                half * y, const int incy );
#endif
  void
  ASPEN_sswap ( const int n,
                float * x, const int incx,
                float * y, const int incy );
  void
  ASPEN_dswap ( const int n,
                double * x, const int incx,
                double * y, const int incy );
  void
  ASPEN_cswap ( const int n,
                cuFloatComplex * x, const int incx,
                cuFloatComplex * y, const int incy );
  void
  ASPEN_zswap ( const int n,
                cuDoubleComplex * x, const int incx,
                cuDoubleComplex * y, const int incy );
  void
  ASPEN_wswap ( const int n,
                cuddreal * x, const int incx,
                cuddreal * y, const int incy );
  void
  ASPEN_uswap ( const int n,
                cuddcomplex * x, const int incx,
                cuddcomplex * y, const int incy );


#if ASPEN_HALF_ENABLED
  void
  ASPEN_hcopy ( const int n,
                const half * x, const int incx,
                half * y, const int incy );
#endif
  void
  ASPEN_scopy ( const int n,
                const float * x, const int incx,
                float * y, const int incy );
  void
  ASPEN_dcopy ( const int n,
                const double * x, const int incx,
                double * y, const int incy );
  void
  ASPEN_ccopy ( const int n,
                const cuFloatComplex * x, const int incx,
                cuFloatComplex * y, const int incy );
  void
  ASPEN_zcopy ( const int n,
                const cuDoubleComplex * x, const int incx,
                cuDoubleComplex * y, const int incy );
  void
  ASPEN_wcopy ( const int n,
                const cuddreal * x, const int incx,
                cuddreal * y, const int incy );
  void
  ASPEN_ucopy ( const int n,
                const cuddcomplex * x, const int incx,
                cuddcomplex * y, const int incy );

#if ASPEN_HALF_ENABLED
  void
  ASPEN_hscal ( const int n,
                const half alpha,
                half * x, const int incx );
#endif
  void
  ASPEN_sscal ( const int n,
                const float alpha,
                float * x, const int incx );
  void
  ASPEN_dscal ( const int n,
                const double alpha,
                double * x, const int incx );
  void
  ASPEN_cscal ( const int n,
                const cuFloatComplex alpha,
                cuFloatComplex * x, const int incx );
  void
  ASPEN_zscal ( const int n,
                const cuDoubleComplex alpha,
                cuDoubleComplex * x, const int incx );
  void
  ASPEN_wscal ( const int n,
                const cuddreal alpha,
                cuddreal * x, const int incx );
  void
  ASPEN_uscal ( const int n,
                const cuddcomplex alpha,
                cuddcomplex * x, const int incx );
  

#if ASPEN_HALF_ENABLED
  void
  ASPEN_HZERO ( half * x, const int n );
#endif
  void
  ASPEN_SZERO ( float * x, const int n );
  void
  ASPEN_DZERO ( double * x, const int n );
  void
  ASPEN_CZERO ( cuFloatComplex * x, const int n );
  void
  ASPEN_ZZERO ( cuDoubleComplex * x, const int n );
  void
  ASPEN_WZERO ( cuddreal * x, const int n );
  void
  ASPEN_UZERO ( cuddcomplex * x, const int n );
 

#ifdef __cplusplus
}
#endif

#define	ASPEN_H_INCLUDED	1
#endif

