#ifndef ASPEN_MAKE_CONST_H_INCLUDED

#include "aspen_types.h"


// generic ----------------------------------
template < class TYPE >
__forceinline__ __host__ __device__ TYPE
makeCONST ( const int x )
{
  return (TYPE)(x);
}

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
makeCONST ( const float x )
{
  return (TYPE)(x);
}

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
makeCONST ( const double x )
{
  return (TYPE)(x);
}

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
makeCONST ( const cuddreal x )
{
  return (TYPE)(x.x);
}

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
makeCONST ( const cuFloatComplex x )
{
  return (TYPE)(x.x);
}

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
makeCONST ( const cuDoubleComplex x )
{
  return (TYPE)(x.x);
}

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
makeCONST ( const cuddcomplex x )
{
  return (TYPE)(x.x.x);
}

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
makeCONST ( const int x, const int y )
{
  return (TYPE)(x);
}

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
makeCONST ( const float x, const float y )
{
  return (TYPE)(x);
}

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
makeCONST ( const double x, const double y )
{
  return (TYPE)(x);
}

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
makeCONST ( const cuddreal x, const cuddreal y )
{
  return (TYPE)(x.x);
}
// generic ----------------------------------


// half ----------------------------------
#if ASPEN_HALF_ENABLED
template < >
__forceinline__ __host__ __device__ half
makeCONST < half > ( const int x )
{
  int    xx = (x > 0) ? (+x) : (-x);
  short  c  = (xx & 0x03ff);
  half   y;
#if CURRENT_GPU==600
  if ( x < 0 ) c |= 0x8000;
#if CUDA_VERSION==8000
  half  z;
#else
  __half_raw  z;
#endif
  z.x = c;
  y = z;
#else
  y = (half)c;
#endif
  return y;
}
#endif
// half ----------------------------------


// cuddreal ----------------------------------
template < >
__forceinline__ __host__ __device__ cuddreal
makeCONST < cuddreal > ( const int x )
{
  cuddreal t = { (double)(x), (double)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddreal
makeCONST < cuddreal > ( const float x )
{
  cuddreal t = { (double)(x), (double)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddreal
makeCONST < cuddreal > ( const double x )
{
  cuddreal t = { (double)(x), (double)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddreal
makeCONST < cuddreal > ( const cuddreal x )
{
  return x;
}

template < >
__forceinline__ __host__ __device__ cuddreal
makeCONST < cuddreal > ( const cuFloatComplex x )
{
  cuddreal t = { (double)(x.x), (double)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddreal
makeCONST < cuddreal > ( const cuDoubleComplex x )
{
  cuddreal t = { (double)(x.x), (double)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddreal
makeCONST < cuddreal > ( const int h, const int l )
{
  cuddreal t = { (double)h, (double)l };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddreal
makeCONST < cuddreal > ( const float h, const float l )
{
  cuddreal t = { (double)h, (double)l };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddreal
makeCONST < cuddreal > ( const double h, const double l )
{
  cuddreal t = { (double)h, (double)l };
  return t;
}
// cuddreal ----------------------------------


// cuFloatComplex ----------------------------------
template < >
__forceinline__ __host__ __device__ cuFloatComplex
makeCONST < cuFloatComplex > ( const int x )
{
  cuFloatComplex t = { (float)(x), (float)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuFloatComplex
makeCONST < cuFloatComplex > ( const float x )
{
  cuFloatComplex t = { (float)(x), (float)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuFloatComplex
makeCONST < cuFloatComplex > ( const double x )
{
  cuFloatComplex t = { (float)(x), (float)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuFloatComplex
makeCONST < cuFloatComplex > ( const cuddreal x )
{
  cuFloatComplex t = { (float)(x.x), (float)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuFloatComplex
makeCONST < cuFloatComplex > ( const cuFloatComplex x )
{
  return x;
}

template < >
__forceinline__ __host__ __device__ cuFloatComplex
makeCONST < cuFloatComplex > ( const cuDoubleComplex x )
{
  cuFloatComplex t = { (float)(x.x), (float)(x.y) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuFloatComplex
makeCONST < cuFloatComplex > ( const cuddcomplex x )
{
  cuFloatComplex t = { (float)(x.x.x), (float)(x.y.x) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuFloatComplex
makeCONST < cuFloatComplex > ( const int x, const int y )
{
  cuFloatComplex t = { (float)(x), (float)(y) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuFloatComplex
makeCONST < cuFloatComplex > ( const float x, const float y )
{
  cuFloatComplex t = { (float)(x), (float)(y) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuFloatComplex
makeCONST < cuFloatComplex > ( const double x, const double y )
{
  cuFloatComplex t = { (float)(x), (float)(y) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuFloatComplex
makeCONST < cuFloatComplex > ( const cuddreal x, const cuddreal y )
{
  cuFloatComplex t = { (float)(x.x), (float)(y.x) };
  return t;
}
// cuFloatComplex ----------------------------------


// cuDoubleComplex ----------------------------------
template < >
__forceinline__ __host__ __device__ cuDoubleComplex
makeCONST < cuDoubleComplex > ( const int x )
{
  cuDoubleComplex t = { (double)(x), (double)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuDoubleComplex
makeCONST < cuDoubleComplex > ( const float x )
{
  cuDoubleComplex t = { (double)(x), (double)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuDoubleComplex
makeCONST < cuDoubleComplex > ( const double x )
{
  cuDoubleComplex t = { (double)(x), (double)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuDoubleComplex
makeCONST < cuDoubleComplex > ( const cuddreal x )
{
  cuDoubleComplex t = { (double)(x.x), (double)(0) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuDoubleComplex
makeCONST < cuDoubleComplex > ( const cuFloatComplex x )
{
  cuDoubleComplex t = { (double)(x.x), (double)(x.y) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuDoubleComplex
makeCONST < cuDoubleComplex > ( const cuDoubleComplex x )
{
  return x;
}

template < >
__forceinline__ __host__ __device__ cuDoubleComplex
makeCONST < cuDoubleComplex > ( const cuddcomplex x )
{
  cuDoubleComplex t = { (double)(x.x.x), (double)(x.y.x) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuDoubleComplex
makeCONST < cuDoubleComplex > ( const int x, const int y )
{
  cuDoubleComplex t = { (double)(x), (double)(y) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuDoubleComplex
makeCONST < cuDoubleComplex > ( const float x, const float y )
{
  cuDoubleComplex t = { (double)(x), (double)(y) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuDoubleComplex
makeCONST < cuDoubleComplex > ( const double x, const double y )
{
  cuDoubleComplex t = { (double)(x), (double)(y) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuDoubleComplex
makeCONST < cuDoubleComplex > ( const cuddreal x, const cuddreal y )
{
  cuDoubleComplex t = { (double)(x.x), (double)(y.x) };
  return t;
}
// cuDoubleComplex ----------------------------------


// cuddcomplex ----------------------------------
template < >
__forceinline__ __host__ __device__ cuddcomplex
makeCONST < cuddcomplex > ( const int x )
{
  cuddcomplex t = { makeCONST<cuddreal>( x ), makeCONST<cuddreal>( 0 ) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddcomplex
makeCONST < cuddcomplex > ( const float x )
{
  cuddcomplex t = { makeCONST<cuddreal>( x ), makeCONST<cuddreal>( 0 ) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddcomplex
makeCONST < cuddcomplex > ( const double x )
{
  cuddcomplex t = { makeCONST<cuddreal>( x ), makeCONST<cuddreal>( 0 ) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddcomplex
makeCONST < cuddcomplex > ( const cuddreal x )
{
  cuddcomplex t = { makeCONST<cuddreal>( x ), makeCONST<cuddreal>( 0 ) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddcomplex
makeCONST < cuddcomplex > ( const cuFloatComplex x )
{
  cuddcomplex t = { makeCONST<cuddreal>( x.x ), makeCONST<cuddreal>( x.y ) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddcomplex
makeCONST < cuddcomplex > ( const cuDoubleComplex x )
{
  cuddcomplex t = { makeCONST<cuddreal>( x.x ), makeCONST<cuddreal>( x.y ) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddcomplex
makeCONST < cuddcomplex > ( const cuddcomplex x )
{
  return x;
}

template < >
__forceinline__ __host__ __device__ cuddcomplex
makeCONST < cuddcomplex > ( const int x, const int y )
{
  cuddcomplex t = { makeCONST<cuddreal>( x ), makeCONST<cuddreal>( y ) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddcomplex
makeCONST < cuddcomplex > ( const float x, const float y )
{
  cuddcomplex t = { makeCONST<cuddreal>( x ), makeCONST<cuddreal>( y ) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddcomplex
makeCONST < cuddcomplex > ( const double x, const double y )
{
  cuddcomplex t = { makeCONST<cuddreal>( x ), makeCONST<cuddreal>( y ) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddcomplex
makeCONST < cuddcomplex > ( const cuddreal x, const cuddreal y )
{
  cuddcomplex t = { x, y };
  return t;
}
// cuddcomplex ----------------------------------


// normalization ----------------------------------
template < class TYPE >
__forceinline__ __host__ __device__ TYPE
Normalize ( const double h, const double l )
{ TYPE r; return r; }

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
Normalize ( const int h, const int l )
{ TYPE r; return r; }

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
Normalize ( const float h, const float l )
{ TYPE r; return r; }

template < class TYPE >
__forceinline__ __host__ __device__ TYPE
Normalize ( const TYPE x )
{ TYPE r; return r; }


template < >
__forceinline__ __host__ __device__ cuddreal
Normalize < cuddreal > ( const double h, const double l )
{
  double hi = h, lo = l;
  hi += l;
  double d = hi - h;
  cuddreal t = { (double)h+lo, (double)(lo-d) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cuddreal
Normalize < cuddreal > ( const int h, const int l )
{
  return Normalize < cuddreal > ( (double)h, (double)l );
}

template < >
__forceinline__ __host__ __device__ cuddreal
Normalize < cuddreal > ( const float h, const float l )
{
  return Normalize < cuddreal > ( (double)h, (double)l );
}

template < >
__forceinline__ __host__ __device__ cuddreal
Normalize < cuddreal > ( const cuddreal x )
{
  return Normalize < cuddreal > ( (double)x.x, (double)x.y );
}

template < >
__forceinline__ __host__ __device__ cudfreal
Normalize < cudfreal > ( const float h, const float l )
{
  float hi = h, lo = l;
  hi += l;
  float d = hi - h;
  cudfreal t = { (float)h+lo, (float)(lo-d) };
  return t;
}

template < >
__forceinline__ __host__ __device__ cudfreal
Normalize < cudfreal > ( const int h, const int l )
{
  return Normalize < cudfreal > ( (float)h, (float)l );
}

template < >
__forceinline__ __host__ __device__ cudfreal
Normalize < cudfreal > ( const double h, const double l )
{
  return Normalize < cudfreal > ( (float)h, (float)l );
}

template < >
__forceinline__ __host__ __device__ cudfreal
Normalize < cudfreal > ( const cudfreal x )
{
  return Normalize < cudfreal > ( (float)x.x, (float)x.y );
}
// normalization ----------------------------------


#define	ASPEN_MAKE_CONST_H_INCLUDED	1
#endif

