#include "../tuning/CURRENT_GPU"
#include "aspen_devel.h"
#include "aspen_types.h"


template < class TYPE, int BLOCK_SIZE, int MULTIPLE >
__global__ void 
API_private(ZERO_kernel) ( TYPE * x, const long n )
{
  long pos = threadIdx.x + MULTIPLE * BLOCK_SIZE * blockIdx.x;
  x += pos;

  if ( pos+BLOCK_SIZE*MULTIPLE <= n ) {
#pragma unroll
    for ( int i=0; i<MULTIPLE; ++i ) {
      x[0] = makeCONST<TYPE> ( 0 );
      if ( MULTIPLE > 1 ) x += BLOCK_SIZE;
    }
  } else {
#pragma unroll
    for ( int i=0; i<MULTIPLE; ++i ) {
      if ( pos+BLOCK_SIZE*i < n ) x[0] = makeCONST<TYPE> ( 0 );
      if ( MULTIPLE > 1 ) x += BLOCK_SIZE;
    }
  }
}


template < class TYPE, int BLOCK_SIZE, int MULTIPLE >
__host__ void
API_private(ZERO_host) ( TYPE * x, const int n )
{
  int t1 = BLOCK_SIZE;
  int tt = MULTIPLE*t1;
  int n_blk = n / tt;
  int n_res = ((n_blk*tt<n)?1:0);
  int g1 = n_blk + n_res;

  int dev = API(get_device_ID)();
  int p_dev; cudaGetDevice( &p_dev );
  if ( dev != p_dev ) cudaSetDevice( dev );

  cudaStream_t stream = API(get_Stream)();

  API_private(ZERO_kernel) < TYPE , BLOCK_SIZE, MULTIPLE >
    <<< g1, t1, 32, stream
    >>> ( x, (long) n );

  if ( dev != p_dev ) cudaSetDevice( p_dev );
}


#define BLOCK_SIZE	(16*12)
#define	MULTIPLE	(2)

#if ASPEN_HALF_ENABLED
extern "C" void
API(HZERO) ( half * x, const int n )
{
  if ( n <= 0 ) return
                  API_private(ZERO_host) < half, BLOCK_SIZE, MULTIPLE > ( x, n );
}
#endif

extern "C" void
API(SZERO) ( float * x, const int n )
{
  if ( n <= 0 ) return
                  API_private(ZERO_host) < float, BLOCK_SIZE, MULTIPLE > ( x, n );
}

extern "C" void
API(DZERO) ( double * x, const int n )
{
  if ( n <= 0 ) return
                  API_private(ZERO_host) < double, BLOCK_SIZE, MULTIPLE > ( x, n );
}

extern "C" void
API(CZERO) ( cuFloatComplex * x, const int n )
{
  if ( n <= 0 ) return
                  API_private(ZERO_host) < cuFloatComplex, BLOCK_SIZE, MULTIPLE > ( x, n );
}

extern "C" void
API(ZZERO) ( cuDoubleComplex * x, const int n )
{
  if ( n <= 0 ) return
                  API_private(ZERO_host) < cuDoubleComplex, BLOCK_SIZE, MULTIPLE > ( x, n );
}

extern "C" void
API(WZERO) ( cuddreal * x, const int n )
{
  if ( n <= 0 ) return
                  API_private(ZERO_host) < cuddreal, BLOCK_SIZE, MULTIPLE > ( x, n );
}

extern "C" void
API(UZERO) ( cuddcomplex * x, const int n )
{
  if ( n <= 0 ) return
                  API_private(ZERO_host) < cuddcomplex, BLOCK_SIZE, MULTIPLE > ( x, n );
}


