#!/bin/sh

echo_Message()
{
	echo '<--/****************************************'
	echo 'Automatic Performance Tuning for DSYMV-L'
	date
	echo -n 'Host on '; echo `hostname`
	echo 'Device is '$DEVICE
	if [ x${1} = xbegin ]; then
		echo 'Start.'
	fi
	if [ x${1} = xend ]; then
		echo 'Successfully completed.'
	fi
	if [ x${1} = xterminate ]; then
		echo 'Terminated unfortunately.'
	fi
	echo '****************************************/-->'
}


export LANG=C
LOG=log-dsymvl-AT1X.`date +%Y.%m.%d.%T`
WORK_DIR=dsymvl-current

if [ x${CUDA_PATH} != x ]; then
        export LD_LIBRARY_PATH=$CUDA_PATH/lib64:$LD_LIBRARY_PATH
fi
export LD_LIBRARY_PATH=`pwd`/../lib:$LD_LIBRARY_PATH

cd ../src; make get_dev_info; cd ../tuning
../src/get_dev_info $ASPEN_GPU_ID > DEV_INFO

DEVICE=`awk '/DEVICE=/{ print $2 }' DEV_INFO`
MP=`awk '/MP=/{ print $2 }' DEV_INFO`
CG=`awk '/CG=/{ print $2 }' DEV_INFO`
MAXDIM=`awk '/MAXDIM=/{ print $2 }' DEV_INFO`


cd ../src; make get_mult get_mult32 get_mult48; cd ../tuning

echo_Message begin

/bin/sh ./touch_header.sh

mkdir $LOG
if [ -f $WORK_DIR ]; then
	\rm -rf $WORK_DIR
fi
if [ -h $WORK_DIR ]; then
	\rm $WORK_DIR
fi
if [ -d $WORK_DIR ]; then
	\rm -rf $WORK_DIR
fi
ln -s $LOG $WORK_DIR

echo 'Tuning processes are recorded on '$LOG'.'

cd $WORK_DIR
cd ../../src; make get_dev_info; cd ../tuning/$WORK_DIR
ln -s ../../src/get_dev_info .

tar -C ../.. -cf - src template | tar -xf -

#cat symv_lower_atomic_template.incl

/bin/sh ../d.sh > OK-pat-symv

touch done_pattern
\rm done_pattern

/bin/sh ../dsymvl_a.sh | tee $LOG
/bin/sh ../dsymvl_b0.sh $LOG
/bin/sh ../dsymvl_b1.sh $LOG-X
/bin/sh ../dsymvl_b2.sh $LOG-X-X
/bin/sh ../dsymvl_b3.sh $LOG-X-X-X
/bin/sh ../dsymvl_c.sh
/bin/sh ../d_filter.sh dsymv-lower-auto3.h dsymv-lower-auto2.h

echo "Complete phase d"

echo '#if 0'             > dsymv-lower-auto_.h
echo_Message            >> dsymv-lower-auto_.h
cat ../DEV_INFO         >> dsymv-lower-auto_.h
echo '<--'              >> dsymv-lower-auto_.h
cat ../CURRENT_GPU      >> dsymv-lower-auto_.h
echo '-->'              >> dsymv-lower-auto_.h
echo '#endif'           >> dsymv-lower-auto_.h
cat dsymv-lower-auto.h  >> dsymv-lower-auto_.h
mv dsymv-lower-auto_.h dsymv-lower-auto.h
cp dsymv-lower-auto.h ..

echo '#if 0'             > dsymv-lower-auto_.h
echo_Message            >> dsymv-lower-auto_.h
cat ../DEV_INFO         >> dsymv-lower-auto_.h
echo '<--'              >> dsymv-lower-auto_.h
cat ../CURRENT_GPU      >> dsymv-lower-auto_.h
echo '-->'              >> dsymv-lower-auto_.h
echo '#endif'           >> dsymv-lower-auto_.h
cat dsymv-lower-auto2.h >> dsymv-lower-auto_.h
mv dsymv-lower-auto_.h dsymv-lower-auto2.h
cp dsymv-lower-auto2.h ..

cat dsymv-lower-auto.h
cat dsymv-lower-auto2.h

cd ../../src

\rm dsymv_lower.cu_o
make

cd ../bench

\rm test-d.o test2-d.o
make 

echo "Complete phase e"

timeout -s KILL 4   ./test-dsymv-l IN-medium >& /dev/null
timeout -s KILL 600 ./test-dsymv-l IN-medium
if [ $? -eq 137 ]; then
timeout -s KILL 600 ./test-dsymv-l IN-medium
fi

echo "Complete phase f1"

timeout -s KILL 4    ./test2-dsymv-l IN-medium >& /dev/null
timeout -s KILL 3600 ./test2-dsymv-l IN-medium
if [ $? -eq 137 ]; then
timeout -s KILL 3600 ./test2-dsymv-l IN-medium
fi

echo "Complete phase f2"

cd ../tuning

touch .done-dsymvl

echo_Message end

exit 0

