#!/bin/bash

export LD_LIBRARY_PATH=../../lib:$LD_LIBRARY_PATH

IN=$1

Make ()
{
	cd ../../src
	\rm uhemv_lower.cu_o
	make
	cd ../bench
	\rm test-u.o test2-u.o
	make 
	cd ../tuning/uhemvl-current
}

Do ()
{
	echo $IN
	awk -f ../anal_symv.awk $IN | \
	tee anal_symv-result | \
	awk ' \
		/N= '$N'/{ \
			print; \
		} \
	' | \
	sort -S 1G -r | \
	head -200 | \
	awk ' \
		BEGIN { \
			N=0; Tmin=1.0; \
		} \
		{ \
			if ( N != $2 ) { \
				N=$2; Tmin=$3; p=1; \
				print $0" "300000; \
			} else { \
			T = (2.0e+3*(1.0*Tmin-$3)*(1.0*Tmin-$3))/(1.0*Tmin*Tmin); \
			T = T * sqrt(1.0*N)/100; \
                        if ( T < 700 ) { \
                           print $0" "(int(100000*exp(-T)+1)); \
                        } else { \
                           print $0" "1; \
                        } } \
			p++; \
		} \
	'
}

Main ()
{
N_PATTERN=`cat IN-exe-a | awk '{ if( i>0 && $1>0 ) printf("%d ",$1); i++; }'`
for N in \
	$N_PATTERN
do
	Do
done
}

Main | \
	awk '\
		{ \
			s=$4" "$5" "$6" "$7" "($8-($8%10)); \
			score[s] += $10; \
		} \
		END{ \
			for(f in score) { \
			if ( f ~ /^[ ]*$/ ) { \
				ff="32 1 1 1 1"; \
				printf("%08d %s\n", 0, ff); \
			} else {
				ff=f; \
				printf("%08d %s\n", score[f], f); \
			} \
			} \
			for(i=0;i<30;i++) { \
				printf("%08d %s\n", 0, ff); \
			} \
		}' | \
	sort -S 1G -r | \
	awk '{ t = $2" "$3" "$4" "$5" "$6; \
		if ( f[t] ) next; f[t] = 1; \
		print $1" "t; } \
		END{ for(i=0;i<20;i++) print "0 0 0 0 0 0"; }' | \
	head -20 | tee top20-2step | \
	awk '\
		BEGIN{ \
			id=1; \
		}\
		{ \
			if ( $1 > 0 ) { \
				score[id]=$1; \
				BLOCK_SIZE[id]=$2; \
				VX[id]=$3; \
				UX[id]=$4; \
				MULTI[id]=$5; \
				MX[id]=$6; \
				id++; \
			} \
		} \
		END { \
			print "{"; \
			print "//=========="; \
			print "if(blk==-1){"; \
			print "\t#include \"uhemv-lower-auto2.h\""; \
			print "}"; \
			print "//=========="; \
			print "#if 0"; \
			print "struct uhemv_auto_tuning_param_t { "; \
			print "\tint\tBLOCK_SIZE, VX, UX, MULTI, MX; "; \
			print "} PARAM["id+1"] = {"; \
				for(i=1;i<id;i++) { \
					print "\t{ "BLOCK_SIZE[i]", "VX[i]", "UX[i]", "MULTI[i]", "MX[i]" },"; \
				} \
			print "\t{ 0, 0, 0, 0, 0 }\n};"; \
			print "#endif"; \
			print "//=========="; \
			print "switch ( BLK ) {"; \
			for(i=1;i<id;i++) { \
			for(m=0;m<100;m+=10) { \
				print "#if KERNEL_"i*100+m; \
				print "case "i*100+m":"; \
				print "\tAPI_private(HESYMVl_ATOMIC__host) < GPU_ARCH, scalar_t, "BLOCK_SIZE[i]", GY2D, "VX[i]", "UX[i]", "MULTI[i]", "m" >"; \
				print "\t\t( n, a, lda, x, incx, y, incy, alpha, beta ); break;"; \
				print "#endif"; \
			} \
			} \
			print "default:"; \
			print "\tbreak;"; \
			print "}"; \
			print "//=========="; \
			print "}"; \
		} \
	' > uhemv-lower-auto.h
cp uhemv-lower-auto.h ../


cat -n top20-2step

touch $IN-X
\rm $IN-X

	ID_max=16
if [ x$ASPEN_TUNING_LEVEL = xROUGH ]; then
	ID_max=10
fi
if [ x$ASPEN_TUNING_LEVEL = xFULL ]; then
	ID_max=20
fi
	ID_list=" 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 "

DONE_KERNEL=0
for ID in \
	$ID_list
do
  if [ $ID -le $ID_max ]; then

  POINT=`awk 'BEGIN{ X=0; }{ if (NR=='$ID' && $1~/[0-9]/) { X=$1;exit; } }END{ print X; }' top20-2step`
  if [ $POINT -gt 100 ]; then

	M00=`awk '{ i++; if (i=='$ID') { print $6; } }' top20-2step`
	M00=`expr $M00 "/" 10`
	M00=`expr $M00 "*" 10`

\rm uhemv-lower-auto2.h
touch uhemv-lower-auto2.h
awk ' \
	BEGIN{ \
		for(i=0;i<100;i++) print "#define\tKERNEL_"'$ID'*100+i"\t" \
			((i-(i%10)) == '$M00'); \
		exit; \
	} \
	' > uhemv-lower-auto2.h
cp uhemv-lower-auto2.h ../

Make >& /dev/null

#for M0 in \
#	0 1 2 3 4 5 6 7 8 9
for M0 in \
	0
do
	BLOCK_SIZE=`awk '{ i++; if (i=='$ID') { print $2; } }' top20-2step`
	VX=`awk '{ i++; if (i=='$ID') { print $3; } }' top20-2step`
	UX=`awk '{ i++; if (i=='$ID') { print $4; } }' top20-2step`
	MULTIPLICITY=`awk '{ i++; if (i=='$ID') { print $5; } }' top20-2step`
	M00=`awk '{ i++; if (i=='$ID') { print $6; } }' top20-2step`
	M00=`expr $M00 "/" 10`
	M00=`expr $M00 "*" 10`
	M11=`expr $M11 "+" 1`
	M=`expr $M0 "+" $M00`

	if [ x$VX != x ]; then
		MM=`awk '{ if ( $1=='$UX'/'$VX' && $2==('$M'%10) ) { print "OK"; } }' OK-pat-symv`
        else
		MM=''
	fi

	if [ x$MM != x ]; then

	WW=`cat log-regs-$BLOCK_SIZE-$VX | \
                awk '/^[0-9]/{ if ( $3=='$UX' && ($4=='$M00' || $4=='$M11') ) { if ( $NF >= '$MULTIPLICITY' && $8 > 0 ) print "OK"; exit; } }'`

        if [ x$WW != x ]; then

cat << EOF | tee -a $IN-X
#define	BLOCK_SIZE	$BLOCK_SIZE
#define	VX		$VX
#define	UX		$UX
#define	MULTIPLICITY	$MULTIPLICITY
#define	_M_		$M
EOF

	KERNEL=`expr $ID "*" 100`
	KERNEL=`expr $KERNEL "+" $M`
	timeout -s KILL 720 ../../bench/test2-uhemv-l IN-exe-a $KERNEL | tee -a $IN-X
	echo $BLOCK_SIZE" "$VX" "$UX" "$MULTIPLICITY" "$M >> done_pattern1
	DONE_KERNEL=`expr $DONE_KERNEL "+" 1`

	fi

	fi
done
  fi

  fi
done

echo "Complete phase b1"
